from .api import run, TablesExecutionError  # exported

## use pypika? ref: https://pypika.readthedocs.io/en/latest/2_tutorial.html
# q = Query.from_('customers').select('id', 'fname', 'lname', 'phone')
# result = run(q.get_sql())



import typing
from dataclasses import dataclass
from .file_utils import download_file


class FileResponse:
    """A file response from the user

    Attributes:
        file (file): The file object
        url (str): The url of the file
        content (bytes): The content of the file
    """

    __file = None

    def __init__(self, url):
        self.url = url
        self.name = url.split("/")[-1]

    @property
    def content(self):
        with open(self.file.name, "rb") as f:
            return f.read()

    @property
    def file(self) -> typing.BinaryIO:
        if not self.__file:
            self.__file = download_file(self.url)
        return self.__file


@dataclass
class PhoneResponse:
    """A phone response from the user

    Attributes:
        masked (str): The masked phone number, eg: +55 (21) 99999-9999
        raw (str): The raw phone number, eg: 5521999999999
        country_code (str): The phone number country code, eg: 55
        national_number (str): The phone number national number, eg: 21999999999
    """

    masked: str
    raw: str
    country_code: str
    national_number: str



class PageResponse(dict):
    def __init__(self, data, action):
        self.action = action
        self.data = data

    def __getattr__(self, key):
        if key in self.data:
            return self.data[key]
        else:
            raise AttributeError(key)

    def get(self, key, default=None):
        return self.data.get(key, default)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __contains__(self, key):
        return key in self.data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __cmp__(self, cmp_dict):
        return self.__cmp__(self.data, cmp_dict)

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    def clear(self):
        return self.data.clear()

    def copy(self):
        return PageResponse(self.data.copy(), self.action)

    def has_key(self, key):
        return key in self.data

    def update(self, *args, **kwargs):
        return self.data.update(*args, **kwargs)

    def pop(self, *args):
        return self.data.pop(*args)



import typing, traceback
from .common import (  # exported
    Message,
    StdioMessage,
    AuthValidJWTMessage,
    AuthInvalidJWTMessage,
    AuthRequireInfoMessage,
    RedirectMessage,
    ExecuteJSRequestMessage,
    AlertMessage,
)
from ..utils import prepate_traceback


class ExecutionIdMessage(Message):
    type = "execution-id"

    def __init__(self, executionId: str):
        super().__init__({"executionId": executionId})


class ErrorMessage(Message):
    type = "error"

    def __init__(self, error: str):
        super().__init__({"error": error})


class ProgramStartFailedMessage(Message):
    type = "program-start-failed"

    def __init__(self, error: str, tb: traceback.StackSummary):
        super().__init__({"error": error, "stack": prepate_traceback(tb)})

    def to_json(self, is_preview: bool):
        if is_preview:
            return super().to_json(is_preview)
        return {"type": self.type}


class EvalReturnMessage(Message):
    type = "eval-return"

    def __init__(self, repr: str):
        super().__init__({"repr": repr})


class EvalErrorMessage(Message):
    type = "eval-error"

    def __init__(self, repr: str):
        super().__init__({"repr": repr})


class AutocompleteSuggestionsMessage(Message):
    type = "autocomplete:suggestions"

    def __init__(self, suggestionsFor: str, suggestions: typing.List[str]):
        super().__init__({"suggestions": suggestions, "suggestionsFor": suggestionsFor})


class WidgetsComputedMessage(Message):
    type = "widgets-computed"

    def __init__(
        self,
        props: dict,
        variables: dict,
        errors: dict,
        seq: int,
    ):
        super().__init__(
            {"props": props, "variables": variables, "errors": errors, "seq": seq}
        )


class WidgetsComputedGeneralErrorMessage(Message):
    type = "widgets-computed"

    def __init__(
        self,
        repr: str,
    ):
        super().__init__({"errors": {"general": {"repr": repr}}})


class SetupRefreshIntervalMessage(Message):
    type = "refresh:setup"

    def __init__(self):
        super().__init__({})



import flask

from ..api import API
from ...usage import usage
from ...settings import Settings
from .utils import send_from_dist
from ...session import StaticSession
from ..runtimes import run_job, run_hook


def get_editor_bp(api: API):
    bp = flask.Blueprint("editor", __name__)

    @bp.route("/api/assets/<path:path>", methods=["GET"])
    def _asset(path):
        return flask.send_from_directory(directory=Settings.root_path, path=path)

    @bp.route("/api/workspace", methods=["GET"])
    @usage
    def get_workspace():
        return api.get_workspace().editor_dto

    @bp.route("/api/workspace", methods=["PUT"])
    @usage
    def update_workspace():
        workspace = api.update_workspace(flask.request.json)
        return workspace.editor_dto

    @bp.route("/api/workspace/root-path", methods=["GET"])
    @usage
    def get_workspace_root_path():
        return str(Settings.root_path.absolute())

    @bp.route("/api/workspace/open-file", methods=["POST"])
    @usage
    def open_file():
        file_path = flask.request.json["path"]
        api.open_file(file_path, create_if_not_exists=True)
        return {"success": True}

    @bp.route("/api/workspace/check-file", methods=["GET"])
    @usage
    def check_file():
        file_path = flask.request.args["path"]
        return {"exists": api.check_file(file_path)}

    @bp.route("/api/workspace/deploy", methods=["POST"])
    @usage
    def deploy():
        api.deploy()
        return {"success": True}

    @bp.route("/api/forms/", methods=["GET"])
    @usage
    def get_forms():
        return [f.editor_dto for f in api.get_forms()]

    @bp.route("/api/forms/<path:path>", methods=["GET"])
    @usage
    def get_form(path: str):
        form = api.get_form(path)
        if not form:
            flask.abort(404)
        return form.editor_dto

    @bp.route("/api/forms/<path:path>", methods=["DELETE"])
    @usage
    def delete_form(path: str):
        api.delete_form(path)
        return {"success": True}

    @bp.route("/api/dashes/<path:path>", methods=["GET"])
    @usage
    def get_dash(path: str):
        dash = api.get_dash(path)
        if not dash:
            flask.abort(404)
        return dash.editor_dto

    @bp.route("/api/forms/", methods=["POST"])
    @usage
    def create_form():
        form = api.create_form()
        return form.editor_dto

    @bp.route("/api/forms/<path:path>", methods=["PUT"])
    @usage
    def update_form(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        form = api.update_form(path, data)
        return form.editor_dto if form else None

    @bp.route("/api/dashes/", methods=["GET"])
    @usage
    def get_dashes():
        return [f.editor_dto for f in api.get_dashes()]

    @bp.route("/api/dashes/", methods=["POST"])
    @usage
    def create_dash():
        dash = api.create_dash()
        return dash.editor_dto

    @bp.route("/api/dashes/<path:path>", methods=["PUT"])
    @usage
    def update_dash(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        dash = api.update_dash(path, data)
        return dash.editor_dto if dash else None

    @bp.route("/api/dashes/<path:path>", methods=["DELETE"])
    @usage
    def delete_dash(path: str):
        api.delete_dash(path)
        return {"success": True}

    @bp.route("/<path:filename>", methods=["GET"])
    @usage
    def spa(filename: str):
        return send_from_dist(filename, "editor.html")

    @bp.route("/", methods=["GET"])
    @usage
    def spa_index():
        return send_from_dist("editor.html", "editor.html")

    @bp.route("/api/hooks/<path:path>", methods=["GET"])
    @usage
    def get_hook(path: str):
        hook = api.get_hook(path)
        if not hook:
            flask.abort(404)
        return hook.editor_dto

    @bp.route("/api/hooks/", methods=["GET"])
    @usage
    def get_hooks():
        return [f.editor_dto for f in api.get_hooks()]

    @bp.route("/api/hooks/", methods=["POST"])
    @usage
    def create_hook():
        hook = api.create_hook()
        return hook.editor_dto

    @bp.route("/api/hooks/<path:path>", methods=["PUT"])
    @usage
    def update_hook(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        hook = api.update_hook(path, data)
        return hook.editor_dto if hook else None

    @bp.route("/api/hooks/<path:path>", methods=["DELETE"])
    @usage
    def delete_hook(path: str):
        api.delete_hook(path)
        return {"success": True}

    @bp.route(
        "/api/hooks/<path:path>/test", methods=["POST", "GET", "PUT", "DELETE", "PATCH"]
    )
    @usage
    def test_hook(path: str):
        hook = api.get_hook(path)
        if not hook:
            flask.abort(404)

        body, status, headers = run_hook(flask.request, hook)
        session = StaticSession.get_session()
        return {
            "body": body,
            "status": status,
            "headers": headers,
            "stdout": "".join(session.stdout if session else []),
            "stderr": "".join(session.stderr if session else []),
        }

    @bp.route("/api/jobs/<path:identifier>", methods=["GET"])
    @usage
    def get_job(identifier: str):
        job = api.get_job(identifier)
        if not job:
            flask.abort(404)
        return job.editor_dto

    @bp.route("/api/jobs/", methods=["GET"])
    @usage
    def get_jobs():
        return [f.editor_dto for f in api.get_jobs()]

    @bp.route("/api/jobs/", methods=["POST"])
    @usage
    def create_job():
        job = api.create_job()
        return job.editor_dto

    @bp.route("/api/jobs/<path:identifier>", methods=["PUT"])
    @usage
    def update_job(identifier: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        job = api.update_job(identifier, data)
        return job.editor_dto if job else None

    @bp.route("/api/jobs/<path:identifier>", methods=["DELETE"])
    @usage
    def delete_job(identifier: str):
        api.delete_job(identifier)
        return {"success": True}

    @bp.route("/api/jobs/<path:identifier>/test", methods=["POST"])
    @usage
    def test_job(identifier: str):
        job = api.get_job(identifier)
        if not job:
            flask.abort(404)

        run_job(job)
        session = StaticSession.get_session()

        return {
            "stdout": "".join(session.stdout if session else []),
            "stderr": "".join(session.stderr if session else []),
        }

    @bp.route("/api/login", methods=["GET"])
    @usage
    def _get_login():
        return api.get_login()

    @bp.route("/api/login", methods=["POST"])
    @usage
    def _create_login():
        data = flask.request.json
        if not data:
            flask.abort(400)

        return api.create_login(data["token"])

    @bp.route("/api/login", methods=["DELETE"])
    @usage
    def _delete_login():
        return api.delete_login()

    @bp.route("/api/ai/message", methods=["POST"])
    @usage
    def _get_next_message():
        body = flask.request.json
        if not body:
            flask.abort(400)

        streamer = api.send_ai_message(body["messages"], body["runtime"])

        if streamer is None:
            flask.abort(403)

        return flask.Response(streamer, mimetype="text/event-stream")

    return bp



import requests, os, typing, abc

from ..credentials import resolve_headers

LOCAL_TABLES_EXECUTE_URL = "https://cloud-api.abstra.cloud/cli/tables/execute"
ABSTRA_TABLES_EXECUTE_URL = os.getenv("ABSTRA_TABLES_EXECUTE_URL")


class TablesApiHttpClient(abc.ABC):
    execute_url: str

    def __init__(self, execute_url: str) -> None:
        self.execute_url = execute_url

    def execute(self, query: str, params: typing.List) -> requests.Response:
        raise NotImplementedError()


class ProductionTablesApiHttpClient(TablesApiHttpClient):
    def execute(self, query: str, params: typing.List) -> requests.Response:
        body = {"query": query, "params": params}
        return requests.post(self.execute_url, json=body)


class LocalTablesApiHttpClient(TablesApiHttpClient):
    def execute(self, query: str, params: typing.List) -> requests.Response:
        body = {"query": query, "params": params}
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to execute a table query")
        return requests.post(self.execute_url, headers=headers, json=body)


def tables_api_http_client_factory() -> TablesApiHttpClient:
    if ABSTRA_TABLES_EXECUTE_URL is None:
        return LocalTablesApiHttpClient(LOCAL_TABLES_EXECUTE_URL)
    else:
        return ProductionTablesApiHttpClient(ABSTRA_TABLES_EXECUTE_URL)



from abc import abstractmethod, ABC
from typing import Any, List, Union


class Widget(ABC):
    @abstractmethod
    def render(self, ctx: dict):
        raise NotImplementedError("render not implemented")

    @abstractmethod
    def set_props(self, props):
        raise NotImplementedError("set_props not implemented")


class Input(Widget):
    type: str
    errors: List[str] = []
    empty_value: Any = None

    def __init__(self, key: str) -> None:
        super().__init__()
        self.key = key

    def has_errors(self):
        return len(self.errors) > 0

    def set_value(self, value, set_errors=False):
        old_value = self.value if hasattr(self, "value") else None
        self.value = value

        if set_errors and old_value != value:
            self.set_errors()

    def set_errors(self):
        self.errors = self.validate()

    def is_value_unset(self):
        if self.has_prop_multiple() and self.multiple:
            return False

        return (
            self.value == None
            or self.value == ""
            or self.value == []
            or (isinstance(self.value, bool) and self.value == False)
            or self.value == {}
        )

    def has_prop_multiple(self):
        return "multiple" in self.__dict__

    def _validate_required(self) -> List[str]:
        if hasattr(self, "required") and self.required and self.is_value_unset():
            if type(self.required) == str:
                return [self.required]

            return ["i18n_error_required_field"]

        return []

    def _validate_number_min_max(self) -> List[str]:
        if type(self.value) != int and type(self.value) != float:
            return []

        if hasattr(self, "min") and self.min is not None and self.value < self.min:
            return ["i18n_error_min_number"]

        if hasattr(self, "max") and self.max is not None and self.value > self.max:
            return ["i18n_error_max_number"]

        return []

    def _validate_list_min_max(self) -> List[str]:
        if type(self.value) != list:
            return []

        if hasattr(self, "min") and self.min is not None and len(self.value) < self.min:
            return ["i18n_error_min_list"]

        if hasattr(self, "max") and self.max is not None and len(self.value) > self.max:
            return ["i18n_error_max_list"]

        return []

    def validate(self) -> List[str]:
        errors = []
        errors.extend(self._validate_required())
        errors.extend(self._validate_number_min_max())
        errors.extend(self._validate_list_min_max())

        return errors

    def __first_or_list(self, value: List) -> Union[List, Any]:
        if not self.has_prop_multiple():
            raise Exception("first_or_list called on widget without multiple property")

        if not isinstance(value, list) and value is not None:
            raise Exception("Non-list value received")

        if value is None:
            return self.empty_value

        if hasattr(self, "multiple") and self.multiple:
            return value if value else []
        return value[0] if len(value) > 0 else None

    @abstractmethod
    def serialize_value(self):
        raise NotImplementedError("serialize_value not implemented")

    def parse_value(self, value):
        _value = value

        if hasattr(self, "multiple"):
            _value = self.__first_or_list(value)

        return _value if _value != None else self.empty_value


class Output(Widget):
    type: str

    def has_errors(self):
        return False



from ..widget_base import Input


class ClickInput(Input):
    type = "click-input"
    value = None

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))
        self.value = None

    def set_props(self, props):
        self.label = props["label"]
        self.hint = props.get("hint", None)
        self.disabled = props.get("disabled", False)
        self.full_width = props.get("full_width", False)
        self.required = props.get("required", False)

    def render(self, context: dict):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "hint": self.hint,
            "disabled": self.disabled,
            "fullWidth": self.full_width,
            "errors": self.errors,
        }

    def serialize_value(self) -> None:
        return None

    def parse_value(self, value: None) -> None:
        return None



from abstra_internals.session import get_live_session_throwable
from abstra_internals.utils import decode_jwt
from abstra_internals.contract import common


class AuthResponse:
    """The response from the authentication process

    Attributes:
      email (str): The email address of the user
    """

    def __init__(self, email: str):
        self.email = email


def get_user():
    session = get_live_session_throwable()
    session.send(common.AuthRequireInfoMessage())

    while True:
        type, data = session.recv()
        if type != "auth:saved-jwt":
            continue

        jwt_claims = decode_jwt(data["jwt"])
        if not jwt_claims:
            session.send(common.AuthInvalidJWTMessage())
            continue

        session.send(common.AuthValidJWTMessage())
        return AuthResponse(jwt_claims["email"])


def execute_js(code: str, context: dict = {}):
    session = get_live_session_throwable()
    session.send(common.ExecuteJSRequestMessage(code, context))

    while True:
        type, data = session.recv()
        if type != "execute-js:response":
            continue

        return data.get("value")


def alert(message: str, severity: str = "info"):
    session = get_live_session_throwable()
    severity = severity if severity in ["info", "warn", "error", "success"] else "info"
    session.send(common.AlertMessage(message, severity))


def redirect(url: str, query_params: dict = {}):
    session = get_live_session_throwable()
    query_params = query_params or get_query_params()
    session.send(common.RedirectMessage(url, query_params))


def get_query_params() -> dict:
    session = get_live_session_throwable()
    return session.context.get("query_params", {})



import fnmatch, re
from pathlib import Path


GIT_FOLDER = ".git"
ABSTRA_FOLDER = ".abstra"
ABSTRA_IGNORE = ".abstraignore"
CREDENTIALS_FILE = ".abstra/credentials"


def get_ignore_files(dir: Path):
    GITIGNORE = dir.joinpath(".gitignore")
    IGNOREFILE = dir.joinpath(ABSTRA_IGNORE)
    git_path = dir.joinpath(GIT_FOLDER)
    abstra_path = dir.joinpath(ABSTRA_FOLDER)
    ignored = [IGNOREFILE.name, GITIGNORE.name, abstra_path.name, git_path.name]

    if IGNOREFILE.exists():
        with open(IGNOREFILE, "r", encoding="utf-8") as f:
            ignored.extend(IGNOREFILE.read_text(encoding="utf-8").split("\n"))
    elif GITIGNORE.exists():
        ignored.extend(GITIGNORE.read_text(encoding="utf-8").split("\n"))
    return [
        re.compile(fnmatch.translate(f"{p}*"))
        for p in ignored
        if p and not p.startswith("#") and not p.startswith("!")
    ]


def should_ignore(ignored_paths, _path: Path):
    path = str(_path)
    for ignored_path in ignored_paths:
        if ignored_path.search(path):
            return True
    return False


def files_from_directory(directory):
    ignored = [*get_ignore_files(directory), *get_ignore_files(Path.cwd())]
    paths = filter(
        lambda p: p.is_file() and not should_ignore(ignored, p.relative_to(directory)),
        Path(directory).rglob("*"),
    )
    return list(paths)



import io
import tempfile
import pathlib
import typing
from .apis import upload_file, internal_path


def convert_file(file: typing.Union[str, io.BufferedReader, io.TextIOWrapper]) -> str:
    if not file:
        return ""

    if isinstance(file, str):
        # URL or base64 encoded string
        if file.startswith("http") or file.startswith("data:"):
            return file

        # path to file
        return upload_file(open(file, "rb"))

    if isinstance(file, io.BufferedReader):
        return upload_file(file)

    if isinstance(file, io.TextIOWrapper):
        return upload_file(file)

    if isinstance(file, pathlib.Path):
        return upload_file(open(file, "rb"))

    # PILImage. TODO: check with isinstance without external dependency
    if hasattr(file, "save"):
        file_path = pathlib.Path(tempfile.gettempdir(), "img.png")
        file.save(str(file_path))
        return upload_file(open(file_path, "rb"))

    # FileResponse. TODO: check with isinstance without circular import
    if hasattr(file, "file"):
        return upload_file(file.file)


def download_file(url: str):
    import requests

    if url.startswith("http://") or url.startswith("https://"):
        f = tempfile.NamedTemporaryFile()
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
        f.seek(0)
        return f

    elif url.startswith("/_files/"):
        name = url[len("/_files/") :]
        path = internal_path(name)
        return open(path, "rb")

    raise ValueError(f"Cannot download {url}")



import sys
from ..session import Execution


def __overload_stdio():
    def writeWraper(type, write, text):
        try:
            write(text)
            execution = Execution.get_execution()
            if execution:
                execution.stdio(type, text)
        finally:
            return len(text)

    stdout_write = sys.stdout.write
    stderr_write = sys.stderr.write

    sys.stdout.write = lambda text: writeWraper("stdout", stdout_write, text)
    sys.stderr.write = lambda text: writeWraper("stderr", stderr_write, text)


def overloads():
    __overload_stdio()



import typing, pathlib, os, sys


class SettingsController(object):
    _instance: typing.Optional["SettingsController"] = None
    _root_path: typing.Optional[pathlib.Path] = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(SettingsController, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def set_root_path(root_path_str: str):
        root_path = pathlib.Path(root_path_str)
        root_path.mkdir(exist_ok=True, parents=True)
        os.chdir(root_path)

        if root_path not in sys.path:
            sys.path.append(str(root_path))

        SettingsController._root_path = (
            (pathlib.Path.cwd() / root_path).resolve().absolute()
        )

    @property
    def root_path(self) -> pathlib.Path:
        if SettingsController._root_path is None:
            raise Exception("You must set the root path before using it")
        return SettingsController._root_path


Settings = SettingsController()



import traceback, flask

from ...api.classes import HookJSON
from ....session import StaticSession


def run_hook(request: flask.Request, hook: HookJSON):
    code = hook.file_path.read_text()
    session = StaticSession("hooks", hook.path)
    session.context["request"] = (
        request.get_data(as_text=True),
        request.args,
        request.headers,
    )

    try:
        namespace: dict = {}
        exec(code, namespace, namespace)
        return session.context.get("response", ({}, 200, {}))
    except:
        traceback.print_exc()
        return session.context.get("response", ({}, 500, {}))



import flask_sock, simple_websocket, typing
from ....session import LiveSession
from ...api.classes import DashJSON
from .dash_runtime import DashRuntime
from ....contract import dashes_contract


def run_dash(
    conn: flask_sock.Server,
    dash: DashJSON,
    session_id: typing.Optional[str] = None,
):
    session = LiveSession(conn, "dashes", dash.path)
    if session_id is not None:
        session.id = session_id

    msg_handler = DashRuntime(session, dash_json=dash)
    session.send(dashes_contract.ExecutionIdMessage(session.id))
    while True:
        try:
            type, data = session.recv()
            msg_handler.handle(type, data)
        except simple_websocket.ConnectionError:
            print("connection closed")



from abc import abstractmethod, ABC


# Represents an Image from PIL package
# from PIL.Image import Image
class PILImage(ABC):
    @abstractmethod
    def save(self, path: str):
        pass


# Represents a DataFrame from pandas package
# from pandas import DataFrame
class PandasDataFrame(ABC):
    @abstractmethod
    def to_json(self, **kwargs):
        pass


# Represents a Figure from plotly package
# from plotly.graph_objects import Figure
class PlotlyFigure(ABC):
    @abstractmethod
    def to_json(self, **kwargs):
        pass



from dataclasses import dataclass
import typing
from .common import (  # exported
    Message,
    StdioMessage,
    AuthValidJWTMessage,
    AuthInvalidJWTMessage,
    AuthRequireInfoMessage,
)


@dataclass
class CloseDTO:
    exit_code: typing.Union[int, None] = None
    exception: typing.Union[str, None] = None


class SessionIdMessage(Message):
    type = "session-id"

    def __init__(self, session_id: str):
        super().__init__({"sessionId": session_id})


class CloseMessage(Message):
    type = "program:end"

    def __init__(self, close_dto: CloseDTO):
        super().__init__(
            {"exitCode": close_dto.exit_code, "exception": close_dto.exception}
        )


class FormUpdateMessage(Message):
    type = "form-update"

    def __init__(self, widgets: typing.List, validation: typing.Dict, event_seq: int):
        super().__init__(
            {
                "widgets": widgets,
                "validation": validation,
                "seq": event_seq,
            }
        )


class FormMessage(Message):
    type = "form"

    def __init__(
        self,
        widgets: typing.List,
        actions: typing.List,
        end_program: bool,
        reactive_polling_interval: int,
        steps: typing.List,
    ):
        super().__init__(
            {
                "widgets": widgets,
                "actions": actions,
                "endProgram": end_program,
                "reactivePollingInterval": reactive_polling_interval,
                "steps": steps,
            }
        )



import typing, threading
from abstra_internals.contract import dashes_contract
from abstra_internals.session import LiveSession, get_live_session_throwable
from .common import get_query_params, get_user, redirect, alert, execute_js  # exported


def __set_refresh_interval(
    session: LiveSession,
    interval: int,
    user_callback: typing.Optional[typing.Callable],
):
    def timer_callback():
        if user_callback:
            user_callback()
        if session.connected:
            session.send(dashes_contract.SetupRefreshIntervalMessage())
            __set_refresh_interval(session, interval, user_callback)

    timer = threading.Timer(interval, timer_callback)
    session.context["refresh_interval"] = timer
    timer.start()


def start_refresh_interval(
    interval: int, user_callback: typing.Optional[typing.Callable] = None
):
    """Experimental"""
    session = get_live_session_throwable()
    __set_refresh_interval(session, interval, user_callback)


def stop_refresh_interval():
    """Experimental"""
    session = get_live_session_throwable()
    timer = session.context.get("refresh_interval")
    if timer:
        timer.cancel()
        session.context["refresh_interval"] = None



import json, os, tempfile, typing, webbrowser
from werkzeug.datastructures import FileStorage
from pathlib import Path


from ...credentials import (
    get_credentials,
    delete_credentials,
    resolve_headers,
    set_credentials,
)
from ...widgets.apis import get_random_filepath, internal_path
from ...cloud_api import get_auth_info, get_ai_messages
from ...settings import Settings
from ...cli.deploy import deploy
from ...utils import random_id
from . import classes


CLOUD_API_ENDPOINT = os.getenv("CLOUD_API_ENDPOINT", "https://cloud-api.abstra.cloud")
ABSTRA_DATABASE_URL = os.environ.get("ABSTRA_DATABASE_URL")


class API:
    def __init__(self):
        self.abstra_json_path = Settings.root_path.joinpath("abstra.json")
        if not self.abstra_json_path.exists():
            self.init_empty()

    def deploy(self):
        deploy()

    def init_empty(self):
        self.persist(classes.AbstraJSON.make_empty())

    def persist(self, abstra_json: classes.AbstraJSON):
        temp_file = Path(tempfile.mkdtemp()) / "abstra.json"

        with temp_file.open("w") as f:
            json.dump(abstra_json.__dict__, f, indent=2)
        temp_file.replace(self.abstra_json_path)

    def __get_abstra_json(self) -> classes.AbstraJSON:
        abstra_json_content = json.loads(
            self.abstra_json_path.read_text(encoding="utf-8")
        )
        return classes.AbstraJSON.from_dict(abstra_json_content)

    def get_workspace(self) -> classes.WorkspaceJSON:
        abstra_json = self.__get_abstra_json()
        return abstra_json.workspace

    def get_page_runtime(
        self, path
    ) -> typing.Union[classes.DashJSON, classes.FormJSON, None]:
        abstra_json = self.__get_abstra_json()

        for form in abstra_json.forms:
            if path == form.path:
                return form

        for dash in abstra_json.dashes:
            if path == dash.path:
                return dash

        return None

    def open_file(self, file_path: str, create_if_not_exists: bool = False):
        complete_file_path = Settings.root_path.joinpath(file_path)

        if (
            create_if_not_exists
            and file_path.endswith(".py")
            and not complete_file_path.is_file()
        ):
            complete_file_path.touch()

        webbrowser.open(complete_file_path.absolute().as_uri())

    def check_file(self, file_path: str):
        return Settings.root_path.joinpath(file_path).is_file()

    def update_workspace(self, changes: typing.Dict[str, typing.Any]):
        abstra_json = self.__get_abstra_json()
        abstra_json.workspace.update(changes, abstra_json.dashes, abstra_json.forms)
        self.persist(abstra_json)
        return abstra_json.workspace

    # Forms CLRUD

    def create_form(self) -> classes.FormJSON:
        abstra_json = self.__get_abstra_json()

        file_name = f"new_form_{random_id()}"
        file = f"{file_name}.py"
        txt = """from abstra.forms import display, read
name = read("What is your name?")
display(f"Hello World, {name}")"""

        Settings.root_path.joinpath(file).write_text(encoding="utf-8", data=txt)
        form = classes.FormJSON(file=file, path=file_name, title="Untitled Form")

        abstra_json.forms.append(form)

        self.persist(abstra_json)

        return form

    def get_forms(self) -> typing.List[classes.FormJSON]:
        abstra_json = self.__get_abstra_json()
        return abstra_json.forms

    def get_form(self, path: str) -> typing.Optional[classes.FormJSON]:
        abstra_json = self.__get_abstra_json()

        for form in abstra_json.forms:
            if path == form.path:
                return form

        return None

    def update_form(
        self, path: str, changes: typing.Dict[str, typing.Any]
    ) -> typing.Optional[classes.FormJSON]:
        abstra_json = self.__get_abstra_json()
        forms = [f for f in abstra_json.forms if f.path == path]
        if len(forms) == 0:
            raise Exception("Form not found for path ", path)
        form = forms[0]
        form.update(changes)

        self.persist(abstra_json)

        return form

    def delete_form(self, path: str):
        abstra_json = self.__get_abstra_json()
        forms = abstra_json.forms
        abstra_json.forms = [f for f in forms if f.path != path]

        self.persist(abstra_json)

    # Dashes CLRUD

    def create_dash(self) -> classes.DashJSON:
        abstra_json = self.__get_abstra_json()

        file_name = f"new_dash_{random_id()}"
        file = f"{file_name}.py"
        txt = "foo = 'bar'"

        Settings.root_path.joinpath(file).write_text(encoding="utf-8", data=txt)
        dash = classes.DashJSON(
            file=file,
            path=f"new-dash_{random_id()}",
            title="Untitled Dash",
            layout=classes.LayoutJSON(
                version="0.2", props={}, slot=classes.SlotJSON({})
            ),
        )

        abstra_json.dashes.append(dash)

        self.persist(abstra_json)

        return dash

    def get_dashes(self) -> typing.List[classes.DashJSON]:
        abstra_json = self.__get_abstra_json()
        return abstra_json.dashes

    def get_dash(self, path: str) -> typing.Optional[classes.DashJSON]:
        abstra_json = self.__get_abstra_json()

        for dash in abstra_json.dashes:
            if path == dash.path:
                return dash

        return None

    def update_dash(
        self, path: str, changes: typing.Dict[str, typing.Any]
    ) -> typing.Optional[classes.DashJSON]:
        abstra_json = self.__get_abstra_json()
        dashes = [d for d in abstra_json.dashes if d.path == path]
        if len(dashes) == 0:
            raise Exception("Dash not found for path ", path)
        dash = dashes[0]
        dash.update(changes)

        self.persist(abstra_json)

        return dash

    def delete_dash(self, path: str):
        abstra_json = self.__get_abstra_json()
        dashes = abstra_json.dashes
        abstra_json.dashes = [d for d in dashes if d.path != path]

        self.persist(abstra_json)

    # Hooks CLRUD

    def create_hook(self) -> classes.HookJSON:
        abstract_json = self.__get_abstra_json()

        file_name = f"new_hook_{random_id()}"
        file = f"{file_name}.py"
        txt = "print('Hello World')"

        Settings.root_path.joinpath(file).write_text(encoding="utf-8", data=txt)
        hook = classes.HookJSON(file=file, path=file_name, title="Untitled Hook")

        abstract_json.hooks.append(hook)

        self.persist(abstract_json)

        return hook

    def get_hooks(self) -> typing.List[classes.HookJSON]:
        abstra_json = self.__get_abstra_json()
        return abstra_json.hooks

    def get_hook(self, path: str) -> typing.Optional[classes.HookJSON]:
        abstra_json = self.__get_abstra_json()

        for hook in abstra_json.hooks:
            if path == hook.path:
                return hook

        return None

    def update_hook(
        self, path: str, changes: typing.Dict[str, typing.Any]
    ) -> typing.Optional[classes.HookJSON]:
        abstra_json = self.__get_abstra_json()
        hooks = [h for h in abstra_json.hooks if h.path == path]
        if len(hooks) == 0:
            raise Exception("Hook not found for path ", path)
        hook = hooks[0]
        hook.update(changes)
        self.persist(abstra_json)

        return hook

    def delete_hook(self, path: str):
        abstra_json = self.__get_abstra_json()
        hooks = abstra_json.hooks
        abstra_json.hooks = [h for h in hooks if h.path != path]

        self.persist(abstra_json)

    # Jobs CLRUD

    def get_jobs(self):
        abstra_json = self.__get_abstra_json()
        return abstra_json.jobs

    def get_job(self, identifier: str):
        abstra_json = self.__get_abstra_json()
        for job in abstra_json.jobs:
            if identifier == job.identifier:
                return job
        return None

    def create_job(self):
        abstra_json = self.__get_abstra_json()

        file_name = f"new_job_{random_id()}"
        file = f"{file_name}.py"
        txt = "print('Hello World')"

        Settings.root_path.joinpath(file).write_text(encoding="utf-8", data=txt)
        job = classes.JobJSON(
            file=file,
            identifier=random_id(),
            schedule="0 * * * *",
            title="Untitled Job",
        )

        abstra_json.jobs.append(job)

        self.persist(abstra_json)

        return job

    def update_job(self, identifier: str, changes: typing.Dict[str, typing.Any]):
        abstra_json = self.__get_abstra_json()
        jobs = [j for j in abstra_json.jobs if j.identifier == identifier]
        if len(jobs) == 0:
            raise Exception("Job not found for identifier ", identifier)
        job = jobs[0]
        job.update(changes)

        self.persist(abstra_json)

        return job

    def delete_job(self, identifier: str):
        abstra_json = self.__get_abstra_json()
        jobs = abstra_json.jobs
        abstra_json.jobs = [j for j in jobs if j.identifier != identifier]

        self.persist(abstra_json)

    # Login

    def get_credentials(self):
        return get_credentials()

    def get_login(self):
        headers = resolve_headers()
        if not headers:
            return {"logged": False, "reason": "NO_API_TOKEN"}
        return get_auth_info(headers)

    def create_login(self, token):
        set_credentials(token)
        return self.get_login()

    def delete_login(self):
        delete_credentials()
        return self.get_login()

    # AI
    def send_ai_message(self, messages, runtime):
        headers = resolve_headers()
        yield from get_ai_messages(messages, runtime, headers)

    # files
    def save_file(self, file: FileStorage, filename: str):
        name, path = get_random_filepath(filename)
        file.save(path)
        return name

    def get_file(self, path: str) -> str:
        return internal_path(path)



import flask_sock, typing, threading, uuid, traceback, typing, threading
from .contract import Message, should_send, StdioMessage
from .utils import serialize, deserialize
from .utils.environment import is_preview
from .monitoring import log, LogMessage


class Execution:
    type = "execution"
    executions: typing.ClassVar[typing.Dict] = {}

    @classmethod
    def get_execution(cls) -> typing.Optional["Execution"]:
        thread_id = threading.get_ident()
        return cls.executions.get(thread_id)

    def __init__(
        self,
        runtime_type: str,
        runtime_name: str,
    ):
        self.id: str = uuid.uuid4().__str__()
        self.thread_id = threading.get_ident()
        self.runtime_type: str = runtime_type
        self.runtime_name: str = runtime_name
        self.stderr: typing.List[str] = []
        self.stdout: typing.List[str] = []
        self.context: typing.Dict = {}
        Execution.executions[self.thread_id] = self

    @property
    def is_preview(self) -> bool:
        return is_preview()

    def stdio(self, type: str, text: str):
        if type == "stderr":
            self.stderr.append(text)
        elif type == "stdout":
            self.stdout.append(text)

    def log(self, event: str, payload: typing.Any):
        if event == "heartbeat":
            return
        log(
            LogMessage(
                event=event,
                payload=payload,
                executionId=self.id,
                runtime_type=self.runtime_type,
                runtime_name=self.runtime_name,
            )
        )

    def delete(self):
        del Execution.executions[self.thread_id]


class LiveSession(Execution):
    type = "session"

    @staticmethod
    def get_session() -> typing.Optional["LiveSession"]:
        execution = Execution.get_execution()
        if isinstance(execution, LiveSession):
            return execution

        return None

    @staticmethod
    def broadcast(msg: Message):
        for e in list(Execution.executions.values()):
            if isinstance(e, LiveSession) and e.connected and e.closed == False:
                e.send(msg)

    def __init__(
        self, connection: flask_sock.Server, runtime_type: str, runtime_name: str
    ):
        super().__init__(runtime_type, runtime_name)
        self.__connection = connection
        self.closed = False

    def send(self, msg: Message):
        self.log(msg.type, msg.data)

        if not should_send(msg, self.is_preview):
            return

        str_data = serialize(msg.to_json(self.is_preview))
        self.__connection.send(str_data)

    def recv(self) -> typing.Tuple[str, typing.Dict]:
        str_data = self.__connection.receive()
        data = deserialize(str_data)
        self.log(data["type"], data)
        return data["type"], data

    def close(self, reason: str = ""):
        self.__connection.close(message=reason)
        self.closed = True

    def stdio(self, type: str, text: str):
        self.send(StdioMessage(type, text))
        return super().stdio(type, text)

    @property
    def connected(self) -> bool:
        return self.__connection.connected

    def end(self):
        if self.connected:
            self.close(reason=traceback.format_exc() or None)
        self.delete()


class StaticSession(Execution):
    type = "static-session"

    @staticmethod
    def get_session() -> typing.Optional["StaticSession"]:
        execution = Execution.get_execution()
        if isinstance(execution, StaticSession):
            return execution

        return None

    def __init__(self, runtime_type: str, runtime_name: str):
        super().__init__(runtime_type, runtime_name)

    def stdio(self, type: str, text: str):
        self.log(type, text)
        return super().stdio(type, text)


def get_live_session_throwable() -> LiveSession:
    session = LiveSession.get_session()
    if not session:
        raise Exception("No session found")
    return session


def get_static_session_throwable() -> StaticSession:
    session = StaticSession.get_session()
    if not session:
        raise Exception("No session found")
    return session


def get_execution_throwable() -> Execution:
    execution = Execution.get_execution()
    if not execution:
        raise Exception("No execution found")
    return execution



from abstra_internals.contract import Message
from abstra_internals.session import get_live_session_throwable


def send(message: Message):
    session = get_live_session_throwable()
    return session.send(message)


def receive():
    session = get_live_session_throwable()
    while True:
        type, data = session.recv()
        if type in ["heartbeat", "browser:try-disconnect"]:
            continue

        return data



import sys
from typing import List, Optional, Union, Any, Dict
from dataclasses import dataclass
from pathlib import Path

from ...utils import check_is_url
from ...settings import Settings


@dataclass
class SidebarRuntime:
    title: str
    path: str


@dataclass
class HookJSON:
    file: str
    path: str
    title: str
    enabled: bool = False

    @property
    def runner_type(self):
        return "hook"

    @property
    def runner_dto(self):
        return {
            "file": self.file,
            "path": self.path,
            "title": self.title,
            "enabled": self.enabled,
        }

    @property
    def editor_dto(self):
        return self.runner_dto

    @property
    def __dict__(self):
        return self.editor_dto

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["path", "title", "enabled"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of hook")

    @staticmethod
    def from_dict(data: dict):
        return HookJSON(
            file=data["file"],
            path=data["path"],
            title=data["title"],
            enabled=data["enabled"],
        )


@dataclass
class JobJSON:
    file: str
    title: str
    identifier: str
    schedule: str

    @property
    def runner_type(self):
        return "job"

    @property
    def runner_dto(self):
        return {
            "file": self.file,
            "identifier": self.identifier,
            "title": self.title,
            "schedule": self.schedule,
        }

    @property
    def editor_dto(self):
        return self.runner_dto

    @property
    def __dict__(self):
        return self.editor_dto

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["identifier", "title", "schedule"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of job")

    @staticmethod
    def from_dict(data: Dict):
        return JobJSON(
            file=data["file"],
            identifier=data["identifier"],
            title=data["title"],
            schedule=data["schedule"],
        )


@dataclass
class FormJSON(SidebarRuntime):
    file: str
    end_message: Optional[str] = None
    auto_start: Optional[bool] = False
    start_message: Optional[str] = None
    error_message: Optional[str] = None
    welcome_title: Optional[str] = None
    allow_restart: Optional[bool] = False
    timeout_message: Optional[str] = None
    start_button_text: Optional[str] = None
    restart_button_text: Optional[str] = None

    @property
    def runner_type(self):
        return "form"

    @property
    def runner_dto(self):
        return {
            "id": self.path,
            "path": self.path,
            "title": self.title,
            "auto_start": self.auto_start,
            "allow_restart": self.allow_restart,
            "end_message": self.end_message,
            "start_message": self.start_message,
            "error_message": self.error_message,
            "welcome_title": self.welcome_title,
            "timeout_message": self.timeout_message,
            "start_button_text": self.start_button_text,
            "restart_button_text": self.restart_button_text,
        }

    @property
    def editor_dto(self):
        return {
            **self.runner_dto,
            "file": self.file,
        }

    @property
    def __dict__(self):
        return {
            "file": self.file,
            "path": self.path,
            "title": self.title,
            "end_message": self.end_message,
            "auto_start": self.auto_start,
            "start_message": self.start_message,
            "error_message": self.error_message,
            "welcome_title": self.welcome_title,
            "allow_restart": self.allow_restart,
            "timeout_message": self.timeout_message,
            "start_button_text": self.start_button_text,
            "restart_button_text": self.restart_button_text,
        }

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in [
                "path",
                "title",
                "end_message",
                "auto_start",
                "start_message",
                "error_message",
                "welcome_title",
                "allow_restart",
                "timeout_message",
                "start_button_text",
                "restart_button_text",
            ]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of form")

    @staticmethod
    def from_dict(data: dict):
        return FormJSON(
            file=data["file"],
            path=data["path"],
            title=data["title"],
            end_message=data["end_message"],
            auto_start=data["auto_start"],
            start_message=data["start_message"],
            error_message=data["error_message"],
            welcome_title=data["welcome_title"],
            allow_restart=data["allow_restart"],
            timeout_message=data["timeout_message"],
            start_button_text=data["start_button_text"],
            restart_button_text=data["restart_button_text"],
        )


def is_widget(x: Dict[str, Any]):
    return "slot" not in x


@dataclass
class DashWidgetJSON:
    id: str
    type: str
    colStart: int
    colEnd: int
    rowStart: int
    rowEnd: int
    props: Dict[str, str]
    events: Dict[str, str]
    variable: Optional[str] = None
    name: Optional[str] = None

    @property
    def runner_dto(self):
        return {
            "id": self.id,
            "type": self.type,
            "name": self.name,
            "variable": self.variable,
            "position": {
                "colStart": self.colStart,
                "colEnd": self.colEnd,
                "rowStart": self.rowStart,
                "rowEnd": self.rowEnd,
            },
            "props": [k for k in self.props.keys()],
            "events": [k for k in self.events.keys()],
        }

    @property
    def editor_dto(self):
        return {
            "id": self.id,
            "type": self.type,
            "variable": self.variable,
            "position": {
                "colStart": self.colStart,
                "colEnd": self.colEnd,
                "rowStart": self.rowStart,
                "rowEnd": self.rowEnd,
            },
            "props": self.props,
            "events": self.events,
        }

    @property
    def __dict__(self):
        return {
            "id": self.id,
            "type": self.type,
            "colStart": self.colStart,
            "colEnd": self.colEnd,
            "rowStart": self.rowStart,
            "rowEnd": self.rowEnd,
            "props": self.props,
            "events": self.events,
            "variable": self.variable,
        }

    def update(self, changes: Dict[str, Any]):
        if "id" in changes:
            self.id = changes["id"]

        if "type" in changes:
            self.type = changes["type"]

        if "variable" in changes:
            self.variable = changes["variable"]

        if "colStart" in changes:
            self.colStart = changes["colStart"]

        if "colEnd" in changes:
            self.colEnd = changes["colEnd"]

        if "rowStart" in changes:
            self.rowStart = changes["rowStart"]

        if "rowEnd" in changes:
            self.rowEnd = changes["rowEnd"]

        if "props" in changes:
            self.props = changes["props"]

        if "events" in changes:
            self.events = changes["events"]

        if "position" in changes:
            self.colStart = changes["position"]["colStart"]
            self.colEnd = changes["position"]["colEnd"]
            self.rowStart = changes["position"]["rowStart"]
            self.rowEnd = changes["position"]["rowEnd"]

    @staticmethod
    def from_dict(obj):
        if "position" in obj:
            obj["colStart"] = obj["position"]["colStart"]
            obj["colEnd"] = obj["position"]["colEnd"]
            obj["rowStart"] = obj["position"]["rowStart"]
            obj["rowEnd"] = obj["position"]["rowEnd"]
            del obj["position"]
        return DashWidgetJSON(**obj)


@dataclass
class SlottableJSON:
    id: str
    type: str
    row: int
    height: int
    order: int
    props: Dict[str, str]
    slot: "SlotJSON"

    def __post_init__(self):
        self.slot = SlotJSON(self.slot)

    @property
    def runner_dto(self):
        return {
            "id": self.id,
            "type": self.type,
            "position": {
                "row": self.row,
                "height": self.height,
                "order": self.order,
            },
            "props": [k for k in self.props.keys()],
            "slot": self.slot.runner_dto,
        }

    @property
    def editor_dto(self):
        return {
            "id": self.id,
            "type": self.type,
            "row": self.row,
            "height": self.height,
            "order": self.order,
            "props": self.props,
            "slot": self.slot.editor_dto,
        }

    @property
    def __dict__(self):
        return {
            "id": self.id,
            "type": self.type,
            "position": {
                "row": self.row,
                "height": self.height,
                "order": self.order,
            },
            "props": self.props,
            "slot": self.slot.editor_dto,
        }

    def update(self, changes: Dict[str, Any]):
        if "id" in changes:
            self.id = changes["id"]

        if "type" in changes:
            self.type = changes["type"]

        if "row" in changes:
            self.row = changes["row"]

        if "height" in changes:
            self.height = changes["height"]

        if "order" in changes:
            self.order = changes["order"]

        if "props" in changes:
            self.props = changes["props"]

        if "slot" in changes:
            self.slot.update(changes["slot"])

        if "position" in changes:
            self.row = changes["position"]["row"]
            self.height = changes["position"]["height"]
            self.order = changes["position"]["order"]

    @staticmethod
    def from_dict(obj):
        if "position" in obj:
            obj["row"] = obj["position"]["row"]
            obj["height"] = obj["position"]["height"]
            obj["order"] = obj["position"]["order"]
            del obj["position"]
        return SlottableJSON(**obj)


class SlotJSON:
    __data: Dict[str, Union[SlottableJSON, DashWidgetJSON]]

    def __init__(self, slot: Dict[str, dict]):
        self.__data = {}
        for id, slottable in slot.items():
            if is_widget(slottable):
                self.__data[id] = DashWidgetJSON.from_dict(slottable)
            else:
                self.__data[id] = SlottableJSON.from_dict(slottable)

    def get(self, id: str):
        if id in self.__data:
            return self.__data[id]

    def items(self):
        return self.__data.items()

    def values(self):
        return self.__data.values()

    @property
    def runner_dto(self):
        return {id: slottable.runner_dto for id, slottable in self.__data.items()}

    @property
    def editor_dto(self):
        return {id: slottable.editor_dto for id, slottable in self.__data.items()}

    @property
    def __dict__(self):
        return {id: slottable.__dict__ for id, slottable in self.__data.items()}

    def update(self, changes: Dict[str, Any]):
        for id, slottable in self.__data.items():
            if id in changes:
                slottable.update(changes[id])


@dataclass
class LayoutJSON:
    version: str
    props: Dict[str, str]
    slot: SlotJSON

    def __post_init__(self):
        self.slot = SlotJSON(self.slot)

    @property
    def runner_dto(self):
        if self.version != "0.2":
            raise ValueError("TODO: convert to 0.2")
        return {
            "props": [k for k in self.props.keys()],
            "version": self.version,
            "slot": self.slot.runner_dto,
        }

    @property
    def editor_dto(self):
        return {
            "props": self.props,
            "version": self.version,
            "slot": self.slot.editor_dto,
        }

    @property
    def __dict__(self):
        return {
            "props": self.props,
            "version": self.version,
            "slot": self.slot.__dict__,
        }


@dataclass
class DashJSON(SidebarRuntime):
    file: str
    layout: LayoutJSON

    def __post_init__(self):
        if not isinstance(self.layout, LayoutJSON):
            self.layout = LayoutJSON(**self.layout)

    @property
    def runner_type(self):
        return "dash"

    @property
    def runner_dto(self):
        return {
            "path": self.path,
            "title": self.title,
            "layout": self.layout.runner_dto,
        }

    @property
    def editor_dto(self):
        return {
            "path": self.path,
            "title": self.title,
            "layout": self.layout.editor_dto,
            "file": self.file,
        }

    @property
    def __dict__(self):
        return {
            "path": self.path,
            "title": self.title,
            "layout": self.layout.__dict__,
            "file": self.file,
        }

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        if "path" in changes:
            self.path = changes["path"]
            del changes["path"]

        if "title" in changes:
            self.title = changes["title"]
            del changes["title"]

        if "layout" in changes:
            self.layout = LayoutJSON(**changes["layout"])
            del changes["layout"]

        if "file" in changes:
            _update_file(self, changes["file"])
            del changes["file"]

        if len(changes) > 0:
            raise Exception("Invalid property update")

    @staticmethod
    def from_dict(data: dict):
        return DashJSON(
            path=data["path"],
            file=data["file"],
            title=data["title"],
            layout=data["layout"],
        )


@dataclass
class SidebarItemJSON:
    id: str
    name: str
    path: str
    type: str
    visible: Optional[bool]

    @property
    def __dict__(self):
        return {
            "id": self.id,
            "name": self.name,
            "path": self.path,
            "type": self.type,
            "visible": self.visible,
        }


@dataclass
class SidebarJSON:
    items: List[SidebarItemJSON]

    @staticmethod
    def from_dict(
        sidebar_data=List, dashes: List[DashJSON] = [], forms: List[FormJSON] = []
    ):
        item_name = lambda path: [s.title for s in [*dashes, *forms] if s.path == path][
            0
        ]
        stored_items = []
        for item in sidebar_data:
            if item["path"] in [d.path for d in dashes] + [f.path for f in forms]:
                item = SidebarItemJSON(
                    id=item["id"],
                    name=item_name(item["path"]),
                    path=item["path"],
                    type=item["type"],
                    visible=item.get("visible"),
                )
                stored_items.append(item)

        for dash in dashes:
            if dash.path not in [item.path for item in stored_items]:
                stored_items.append(
                    SidebarItemJSON(
                        id=dash.path,
                        name=dash.title,
                        path=dash.path,
                        type="dash",
                        visible=False,
                    )
                )

        for form in forms:
            if form.path not in [item.path for item in stored_items]:
                stored_items.append(
                    SidebarItemJSON(
                        id=form.path,
                        name=form.title,
                        path=form.path,
                        type="form",
                        visible=False,
                    )
                )
        return SidebarJSON(items=stored_items)

    @property
    def __dict__(self):
        return [item.__dict__ for item in self.items]


@dataclass
class WorkspaceJSON:
    name: str
    sidebar: SidebarJSON
    root: Optional[str] = None  # deprecated
    theme: Optional[str] = None
    logo_url: Optional[str] = None
    brand_name: Optional[str] = None
    main_color: Optional[str] = None
    font_family: Optional[str] = None
    font_color: Optional[str] = None

    @property
    def __dict__(self):
        return {
            "name": self.name,
            "sidebar": self.sidebar.__dict__,
            "root": self.root,  # deprecated
            "theme": self.theme,
            "logo_url": self.logo_url,
            "brand_name": self.brand_name,
            "main_color": self.main_color,
            "font_family": self.font_family,
            "font_color": self.font_color,
        }

    @property
    def runner_dto(self):
        if isinstance(self.logo_url, str) and check_is_url(self.logo_url):
            logo_url = self.logo_url
        elif self.logo_url:
            logo_url = "/_assets/logo"
        else:
            logo_url = None
        return {**self.__dict__, "logo_url": logo_url}

    @property
    def editor_dto(self):
        return self.__dict__

    def update(
        self,
        changes: Dict[str, Any],
        dashes: List[DashJSON] = [],
        forms: List[FormJSON] = [],
    ):
        for attr, value in changes.items():
            if attr == "sidebar":
                self.sidebar = SidebarJSON.from_dict(changes["sidebar"], dashes, forms)
            else:
                setattr(self, attr, value)

    @staticmethod
    def from_dict(data: Dict, dashes: List[DashJSON] = [], forms: List[FormJSON] = []):
        return WorkspaceJSON(
            name=data.get("name", "Untitled Workspace"),
            sidebar=SidebarJSON.from_dict(
                data.get("sidebar", []),
                dashes=dashes,
                forms=forms,
            ),
            root=data.get("root"),  # deprecated
            theme=data.get("theme"),
            logo_url=data.get("logo_url"),
            brand_name=data.get("brand_name"),
            main_color=data.get("main_color"),
            font_family=data.get("font_family"),
            font_color=data.get("font_color"),
        )


@dataclass
class AbstraJSON:
    version: str
    workspace: WorkspaceJSON
    forms: List[FormJSON]
    dashes: List[DashJSON]
    hooks: List[HookJSON]
    jobs: List[JobJSON]

    def __post_init__(self):
        self.forms = [FormJSON.from_dict(form) for form in self.forms]
        self.dashes = [DashJSON.from_dict(dash) for dash in self.dashes]
        self.hooks = [HookJSON.from_dict(hook) for hook in self.hooks]
        self.jobs = [JobJSON.from_dict(job) for job in self.jobs]
        self.workspace = WorkspaceJSON.from_dict(
            self.workspace, dashes=self.dashes, forms=self.forms
        )

    @property
    def __dict__(self):
        return {
            "version": self.version,
            "workspace": self.workspace.__dict__,
            "forms": [form.__dict__ for form in self.forms],
            "dashes": [dash.__dict__ for dash in self.dashes],
            "hooks": [hook.__dict__ for hook in self.hooks],
            "jobs": [job.__dict__ for job in self.jobs],
        }

    @staticmethod
    def from_dict(data: dict):
        if "hooks" not in data:
            data["hooks"] = []

        if "jobs" not in data:
            data["jobs"] = []

        if "forms" not in data:
            data["forms"] = []

        if "dashes" not in data:
            data["dashes"] = []

        if "workspace" not in data:
            data["workspace"] = {
                "name": "Untitled Workspace",
                "sidebar": [],
            }

        try:
            return AbstraJSON(**data)
        except TypeError as e:
            print("Error: incompatible abstra.json file.")
            print(e)
            sys.exit(1)

    @staticmethod
    def make_empty():
        return AbstraJSON(
            version="0.1",
            workspace=dict(
                name="Untitled Workspace",
                sidebar=[],
            ),
            forms=[],
            dashes=[],
            hooks=[],
            jobs=[],
        )


def _update_file(
    runtime: Union[DashJSON, FormJSON, HookJSON, JobJSON],
    new_file_relative: str,
):
    old_file = Settings.root_path.joinpath(runtime.file)
    new_file = Settings.root_path.joinpath(new_file_relative)

    if old_file.exists() and not new_file.exists():
        old_file.rename(new_file)

    runtime.file = new_file_relative



import flask, flask_cors, pathlib

from .api import API
from .overloads import overloads
from .apps.player import get_player_bp
from ..settings import SettingsController


# USED ON CLOUD SERVER
def get_cloud_app(root: str):
    overloads()
    SettingsController.set_root_path(root)

    api = API()
    app = flask.Flask(__name__)
    flask_cors.CORS(app)

    player = get_player_bp(api)
    app.register_blueprint(player)

    return app



from .connectors import connectors_api_http_client_factory
from .tables import tables_api_http_client_factory

connectors_api_http_client = connectors_api_http_client_factory()
tables_api_http_client = tables_api_http_client_factory()
